#!/usr/local/bin/python
#
# $Id: //projects/webval/webval.py#5 $ $Date: 2003/08/23 $

"""
A URL scanner and validator.
"""

__program__ = 'webval'
__version__ = '1.0.1'
__url__ = 'http://www.alcyone.com/pyos/webval/'
__author__ = 'Erik Max Francis <max@alcyone.com>'
__copyright__ = 'Copyright (C) 2002 Erik Max Francis'
__license__ = 'GPL'


import getopt
import httplib
import os
import re
import socket
import sys
import threading
import time


try:
    HOME_DIRECTORY = os.environ['HOME']
except KeyError:
    HOME_DIRECTORY = '.'

AGENT = 'WebVal/' + __version__
IGNORE = []
DATABASE_FILENAME = HOME_DIRECTORY + '/.webval.db'
METHOD = 'HEAD'
JOBS = 10
BAD_CODE = 300
CODES = 5
THRESHHOLD = CODES
TIMEOUT = 2*60
VERBOSE = 1


class Record:
    def __init__(self, url, touched=None, validated=0, codes=None):
        if url.find(' ') >= 0:
            url 
        self.url = url
        if touched is None:
            self.touched = int(time.time())
        else:
            self.touched = touched
        self.validated = validated
        if codes is None:
            self.codes = []
        else:
            self.codes = codes

    def bad(self):
        badCodes = filter(lambda x: x >= BAD_CODE, self.codes)
        if len(badCodes) >= THRESHHOLD:
            return 1
        else:
            return 0

    def touch(self):
        self.touched = int(time.time())

    def validate(self, code):
        self.validated = int(time.time())
        self.codes.append(code)
        if len(self.codes) > CODES:
            self.codes = self.codes[-CODES:]


class Worker(threading.Thread):
    URL_RE = re.compile(r'^http://([^/]+)(/?.*)$')
    
    def __init__(self, url):
        threading.Thread.__init__(self)
        self.setDaemon(1)
        self.url = url
        self.code = None
        self.when = time.time()
        
    def run(self):
        match = self.URL_RE.search(self.url)
        assert match is not None, match.groups()
        groups = match.groups()
        assert len(groups) == 2
        host, file = groups
        if not file:
            file = '/'
        try:
            connection = httplib.HTTPConnection(host)
            connection.request(METHOD, file, \
                               headers={'User-Agent': AGENT})
            response = connection.getresponse()
            self.code = response.status
            connection.close()
        except socket.error: # socket error
            self.code = 900
        except httplib.HTTPException: # httplib error
            self.code = 901
        except IOError: # TCP/IP IO error
            self.code = 902
        except ValueError: # int(port) failed (bad URL)
            self.code = 903


class Database(dict):
    HTTP_RE = re.compile(r'http://[^ \v\r\t\n\'\"<>,\(\)]+')

    def __init__(self, loadFile=None):
        dict.__init__(self)
        if loadFile is not None:
            self.load(loadFile)

    def add(self, record):
        self[record.url] = record

    def save(self, file):
        keys = self.keys()
        keys.sort()
        for key in keys:
            record = self[key]
            if record.codes:
                codes = ':'.join(map(str, record.codes))
            else:
                codes = ':'
            file.write("%s %d %d %s\n" % (record.url, record.touched, \
                                          record.validated, codes))

    def load(self, file):
        for line in file.xreadlines():
            line = line[:-1]
            url, touchedS, validatedS, codesS = line.split(' ', 3)
            touched = int(touchedS)
            validated = int(validatedS)
            if codesS == ':':
                codes = []
            else:
                codes = map(int, codesS.split(':'))
            assert not self.has_key(url)
            self.add(Record(url, touched, validated, codes))

    def touch(self, url):
        """Declare that you've seen this URL."""
        if self.has_key(url):
            self[url].touch()
        else:
            self.add(Record(url))

    def scan(self, file, name=None):
        """Scan a file for URLs and add/touch them."""
        lineNumber = 0
        for line in file.xreadlines():
            lineNumber += 1
            matches = self.HTTP_RE.findall(line)
            for url in matches:
                for ignore in IGNORE:
                    if url.find(ignore) >= 0:
                        break
                else:
                    self.touch(url)
                    if name is not None:
                        record = self[url]
                        if record.bad():
                            codesS = ' '.join(map(str, record.codes))
                            print >> sys.stderr, "%s:%d: %s (%s)" % \
                                  (name, lineNumber, url, codesS)

    def validate(self):
        """Validate the entries in the database."""
        urls = self.keys()
        pending = {}
        while urls or pending:
            if VERBOSE:
                print '... %d left' % (len(urls) + len(pending)) ###
            # Scan through for finished threads.
            for url, thread in pending.items():
                if not thread.isAlive():
                    code = thread.code
                    assert code is not None
                    record = self[url]
                    record.validate(code)
                    del pending[url]
                    if VERBOSE:
                        print code, url
            # If there are empty slots, start new threads.
            while len(pending) < JOBS and urls:
                url = urls.pop()
                thread = Worker(url)
                pending[url] = thread
                thread.start()
            # Pause briefly.
            time.sleep(1.0)
            # Check for the timeout case.
            for url, thread in pending.items():
                if thread.when + TIMEOUT <= time.time():
                    for url in pending.keys():
                        code = 999
                        record = self[url]
                        record.validate(code)
                        del pending[url]
                        if VERBOSE:
                            print code, url


def load():
    """Load a database, or create a new one from scratch if none exists."""
    try:
        file = open(DATABASE_FILENAME, 'r')
        return Database(file)
    except IOError:
        print >> sys.stderr, "no database; starting fresh"
        return Database()

def save(database):
    """Save a database."""
    NEW_DATABASE_FILENAME = DATABASE_FILENAME + '.new'
    file = open(NEW_DATABASE_FILENAME, 'w')
    database.save(file)
    os.rename(NEW_DATABASE_FILENAME, DATABASE_FILENAME)

def usage():
    print >> sys.stderr, """\
Usage: %s [options] <mode> <filenames, or '-' for stdin...>

Modes:
  -s --scan                  Scan files for URLs [default]
  -r --report                Scan files for URLs and report bad ones
  -v --validate              Validate all URLs in database

General options:
  -V --version               Print version and exit
  -h --help                  Print help and exit
  -d --database=<filename>   Specify database file to use [~/.webval.db]

Scanning and reporting options:
  -b --bad-code=<code>       The lowest code which is considered bad [%d]
  -t --treshhold=<count>     Number of bad codes before reporting error [%d]
  -I --ignore=<strings>      Comma separated list of URL fragments to skip

Validation options:
  -q --quiet                 Do not print status while validating
  -k --codes=<count>         Number of validation results to keep [%d]
  -j --jobs=<threads>        Number of threads to use while validating [%d]
  -A --agent=<agent>         Agent string to use in HTTP requests [%s]
  -M --method=<type>         HTTP method to use (HEAD or GET) [%s]
  -T --timeout=<seconds>     HTTP request timeout [%d]

Codes are normal HTTP codes (2xx = success, 3xx = redirect, 4xx = error,
5xx = server error), with 9xx being a special extension used to indicate
metaerrors (DNS lookup failures, malformed URLs, server timeouts).  Any code
greater or equal to 300 is considered bad.
""" % (sys.argv[0], BAD_CODE, THRESHHOLD, CODES, JOBS, AGENT, METHOD, TIMEOUT),

def main():
    global DATABASE_FILENAME, BAD_CODE, THRESHHOLD, CODES, JOBS, AGENT, IGNORE, METHOD, TIMEOUT, VERBOSE
    # Initialize options.
    mode = 'scan'
    # Parse the command line.
    pairs, remainder = getopt.getopt(sys.argv[1:], 'vVhqsrVd:b:t:k:j:A:I:M:T:', ['version', 'help', 'quiet', 'scan', 'report', 'validate', 'database=', 'bad-code=', 'threshold=', 'threshhold=', 'codes=', 'jobs=', 'agent=', 'ignore=', 'method=', 'timeout='])
    for option, argument in pairs:
        if option in ('-V', '--version'):
            sys.stderr.write("%s version %s\n" % (__program__, __version__))
            return
        elif option in ('-h', '--help'):
            usage()
            return
        elif option in ('-q', '--quiet'):
            VERBOSE = 0
        elif option in ('-s', '--scan'):
            mode = 'scan'
        elif option in ('-r', '--report'):
            mode = 'report'
        elif option in ('-v', '--validate'):
            mode = 'validate'
        elif option in ('-d', '--database'):
            DATABASE_FILENAME = argument
        elif option in ('-b', '--bad-code'):
            BAD_CODE = int(argument)
            if BAD_CODE <= 0:
                raise ValueError, "bad code must be positive"
        elif option in ('-t', '--threshold', '--threshhold'):
            THRESHHOLD = int(argument)
            if THRESHHOLD <= 0:
                raise ValueError, "threshhold must be positive"
        elif option in ('-k', '--codes'):
            CODES = int(argument)
            if CODES <= 0:
                raise ValueError, "codes must be positive"
        elif option in ('-j', '--jobs'):
            JOBS = int(jobs)
            if JOBS <= 0:
                raise ValueError, "jobs must be positive"
        elif option in ('-A', '--agent'):
            AGENT = argument
        elif option in ('-I', '--ignore'):
            IGNORE = argument.split(',')
        elif option in ('-M', '--method'):
            METHOD = argument
        elif option in ('-T', '--timeout'):
            TIMEOUT = float(argument)
    if not remainder:
        remainder.append('-')
    database = load()
    if mode == 'scan':
        for filename in remainder:
            if filename == '-':
                file = sys.stdin
            else:
                file = open(filename, 'r')
            database.scan(file)
            file.close()
    elif mode == 'report':
        for filename in remainder:
            if filename == '-':
                file = sys.stdin
            else:
                file = open(filename, 'r')
            database.scan(file, filename)
            file.close()
    elif mode == 'validate':
        database.validate()
    else:
        raise ValueError, "unknown mode"
    save(database)
    
if __name__ == '__main__': main()
